/*

by Luigi Auriemma

*/

#define QUAKE3_QUERY    "\xff\xff\xff\xff" "getinfo xxx\n"
#define MOH_QUERY       "\xff\xff\xff\xff" "\x02" "getinfo xxx\n"
#define QUAKE2_QUERY    "\xff\xff\xff\xff" "status"
#define HALFLIFE_QUERY  "\xff\xff\xff\xff" "infostring\n\0"
#define DPLAY8_QUERY    "\x00\x02"          /* info/join */ \
                        "\xff\xff"          /* ID tracker */ \
                        "\x02"              /* 01 = need GUID, 02 = no GUID */  
#define DOOM3_QUERY     "\xff\xff" "getInfo\0" "\0\0\0\0"
#define ASE_QUERY       "s"



void generic_query(int nt, u_char chr, u_char *data, int len, int front, int rear) {
    u_char  *p,
            *limit;

    limit = data + len - rear;
    data += front;
    while((data < limit) && (p = strchr(data, chr))) {
        *p = 0x00;
        if(nt & 1) {
            fprintf(fdout, "%s\n", data);
        } else {
            if(!*data) break;
            fprintf(fdout, "%30s: ", data);
        }
        data = p + 1;
        nt++;
    }
    fprintf(fdout, "%s\n\n", data);
}



void show_unicode(u_char *ptr, u_int size) {
    u_char  *limit = ptr + size;

    while(*ptr && (ptr <= limit)) {
        fputc(*ptr++, fdout);
        ptr++;
    }
    fputc('\n', fdout);
}



void dplay8info(u_char *data, int len) {
    struct myguid {
        u_int   g1;
        u_short g2;
        u_short g3;
        u_char  g4;
        u_char  g5;
        u_char  g6;
        u_char  g7;
        u_char  g8;
        u_char  g9;
        u_char  g10;
        u_char  g11;
    };
    struct dplay8_struct {
        u_int   pcklen;
        u_int   reservedlen;
        u_int   flags;
        u_int   session;
        u_int   max_players;
        u_int   players;
        u_int   fulldatalen;
        u_int   desclen;
        u_int   dunno1;
        u_int   dunno2;
        u_int   dunno3;
        u_int   dunno4;
        u_int   datalen;
        u_int   appreservedlen;
        struct  myguid  instance_guid;
        struct  myguid  application_guid;
    } *dplay8;

    if(len < (sizeof(struct dplay8_struct) + 4)) {
        fprintf(fdout, "\nError: the packet received seems invalid\n");
        return;
    }

    dplay8 = (struct dplay8_struct *)(data + 4);

    if(dplay8->session) {
        fputs("\nSession options:     ", fdout);
        if(dplay8->session & 1) fputs("client-server ", fdout);
        if(dplay8->session & 4) fputs("migrate_host ", fdout);
        if(dplay8->session & 64) fputs("no_DPN_server ", fdout);
        if(dplay8->session & 128) fputs("password ", fdout);
    }

    fprintf(fdout, "\n"
        "Max players          %u\n"
        "Current players      %u\n",
        dplay8->max_players,
        dplay8->players);

    fprintf(fdout, "\n"
        "Instance GUID        %08x-%04hx-%04hx-%02hhx%02hhx-%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx",
        dplay8->instance_guid.g1,
        dplay8->instance_guid.g2,
        dplay8->instance_guid.g3,
        dplay8->instance_guid.g4,
        dplay8->instance_guid.g5,
        dplay8->instance_guid.g6,
        dplay8->instance_guid.g7,
        dplay8->instance_guid.g8,
        dplay8->instance_guid.g9,
        dplay8->instance_guid.g10,
        dplay8->instance_guid.g11);

    fprintf(fdout, "\n"
        "Application GUID     %08x-%04hx-%04hx-%02hhx%02hhx-%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx\n",
        dplay8->application_guid.g1,
        dplay8->application_guid.g2,
        dplay8->application_guid.g3,
        dplay8->application_guid.g4,
        dplay8->application_guid.g5,
        dplay8->application_guid.g6,
        dplay8->application_guid.g7,
        dplay8->application_guid.g8,
        dplay8->application_guid.g9,
        dplay8->application_guid.g10,
        dplay8->application_guid.g11);

        // removed any security check here

    fputs("\nGame description/Session name:\n\n  ", fdout);
    show_unicode(
        data + 4 + dplay8->fulldatalen,
        len - dplay8->fulldatalen);

    if(dplay8->appreservedlen) {
        fprintf(fdout, "\nHexdump of the APPLICATION RESERVED data (%u) sent by the server:\n\n",
            dplay8->appreservedlen);
        show_dump(
            data + 4 + dplay8->datalen,
            dplay8->appreservedlen,
            fdout);
    }

    if(dplay8->reservedlen) {
        fprintf(fdout, "\nHexdump of the RESERVED data (%u) sent by the server:\n\n",
            dplay8->reservedlen);
        show_dump(
            data + 4 + dplay8->fulldatalen + dplay8->desclen,
            dplay8->reservedlen,
            fdout);
    }

    fputc('\n', fdout);
}



void ase_query(u_char *data, u_int size) {
    u_char  *limit;
    int     num = 0;

    if(memcmp(data, "EYE1", 4)) {
        fwrite(data + 1, size - 1, 1, stdout);
        fputc('\n', stdout);
        return;
    }
    limit = data + size;
    data += 4;

    while((data < limit) && (*data > 1)) {
        switch(num) {
            case 0: fputs("\nRunning game:   ", stdout); break;
            case 1: fputs("\nGame port:      ", stdout); break;
            case 2: fputs("\nServer name:    ", stdout); break;
            case 3: fputs("\nBattle mode:    ", stdout); break;
            case 4: fputs("\nMap name:       ", stdout); break;
            case 5: fputs("\nVersion:        ", stdout); break;
            case 6: fputs("\nPassword:       ", stdout); break;
            case 7: fputs("\nPlayers:        ", stdout); break;
            case 8: fputs("\nMax players:    ", stdout); break;
            default: {
                if(num & 1) fputc('\n', stdout);
                  else fputs(":   ", stdout);
                } break;
        }
        fwrite(data + 1, *data - 1, 1, stdout);
        data += *data;
        num++;
    }

    fputc('\n', stdout);
    num = limit - data;
    if(num > 1) {
        fputs("\nHex dump of the rest of the packet:\n", stdout);
        show_dump(data, num, stdout);
    }
    fputc('\n', stdout);
}



void multi_query(int type, u_int ip, u_short port) {
    struct  sockaddr_in peer;
    int     sd,
            len,
            querylen;
    u_char  buff[4096],
            *query;

#define QUERY_ASSIGN(x,y)   { \
                                fputs("\nQuery: " x "\n", fdout); \
                                query = y; \
                                querylen = sizeof(y) - 1; \
                            }
    switch(type) {
        case 1: QUERY_ASSIGN("Quake 3 engine", QUAKE3_QUERY);   break;
        case 2: QUERY_ASSIGN("Medal of Honor", MOH_QUERY);      break;
        case 3: QUERY_ASSIGN("Quake 2 engine", QUAKE2_QUERY);   break;
        case 4: QUERY_ASSIGN("Half-Life",      HALFLIFE_QUERY); break;
        case 5: QUERY_ASSIGN("DirectPlay 8",   DPLAY8_QUERY);   break;
        case 6: QUERY_ASSIGN("Doom 3 engine",  DOOM3_QUERY);    break;
        case 7: QUERY_ASSIGN("All-seeing-eye", ASE_QUERY);      break;
        default: {
            fputs("\nError: you have choosen an invalid type of query\n\n", stderr);
            exit(1);
            }
    }
#undef QUERY_ASSIGN

    peer.sin_addr.s_addr = ip;
    peer.sin_port        = htons(port);
    peer.sin_family      = AF_INET;

    fprintf(fdout, "Target   %s : %hu\n",
        inet_ntoa(peer.sin_addr), port);

    sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(sd < 0) std_err();

    if(sendto(sd, query, querylen, 0, (struct sockaddr *)&peer, sizeof(peer))
      < 0) std_err();

    if(timeout(sd) < 0) {
        fputs("\n"
            "Error: Socket timeout, no reply received\n"
            "\n", stderr);
        exit(1);
    }
    len = recvfrom(sd, buff, sizeof(buff) - 1, 0, NULL, NULL);
    if(len < 0) std_err();
    buff[len] = 0x00;

    switch(type) {
        case 1: generic_query(1, '\\', buff, len, 4, 0); break;
        case 2: generic_query(1, '\\', buff, len, 5, 0); break;
        case 3: generic_query(1, '\\', buff, len, 4, 0); break;
        case 4: generic_query(1, '\\', buff, len, 23, 0); break;
        case 5: dplay8info(buff, len); break;
        case 6: generic_query(0, '\0', buff, len, 23, 8); break;
        case 7: ase_query(buff, len); break;
        default: break;
    }
    close(sd);
}


